# Generated by Django 2.2.12 on 2020-04-08 17:09

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0006_opensuse_support'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmpublication',
            name='metadata_checksum_type',
            field=models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], default='sha256', max_length=10),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='rpmpublication',
            name='package_checksum_type',
            field=models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], default='sha256', max_length=10),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='original_checksum_types',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]
