# Generated by Django 4.2.7 on 2023-11-22 18:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("rpm", "0055_add_repo_config_field"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="rpmpublication",
            name="sqlite_metadata",
        ),
        migrations.RemoveField(
            model_name="rpmrepository",
            name="sqlite_metadata",
        ),
    ]
