# Generated by Django 4.2.5 on 2023-11-07 03:51

from django.db import migrations, models


def set_publication_checksum(apps, schema_editor):
    RpmPublication = apps.get_model("rpm", "RpmPublication")
    RpmPublication.objects.filter(checksum_type="").update(checksum_type="unknown")


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0059_rpmpublication_compression_type_and_more'),
    ]

    operations = [
        migrations.RunPython(set_publication_checksum),
    ]
