# Generated by Django 3.2.13 on 2022-05-02 07:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0041_modulemdobsolete'),
    ]

    operations = [
        migrations.AlterField(
            model_name='repometadatafile',
            name='data_type',
            field=models.TextField(),
        ),
    ]
